//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "main.h"
#include "plot.h"
#include "plotdlg.h"
#include "graph.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TSolPlot *SolPlot;
//---------------------------------------------------------------------------
#define CHARDEG     0xB0               // deg-character code
#define INTARROW    60.0               // direction arrow interval (sec)

const static TColor MColor[]={         // mark colors by q-flags
	clGray,clGreen,(TColor)0x00AAFF,clGray,clBlue,clRed,clGray
};
TColor TSolPlot::CColor[4]={clWhite,clGray,clBlack,clBlue};
int TSolPlot::PlotTypeD,TSolPlot::QFlagD,TSolPlot::PlotStyle;
int TSolPlot::MarkSize=3,TSolPlot::Origin,TSolPlot::TimeLabel;
int TSolPlot::AutoScale,TSolPlot::ShowArrow,TSolPlot::ShowStats;
double TSolPlot::YRange,TSolPlot::OOPos[3];
//---------------------------------------------------------------------------
__fastcall TSolPlot::TSolPlot(TComponent* Owner) : TForm(Owner)
{
	Drag=NPos[0]=NPos[1]=NPos[2]=0; Xn=Yn=-1;
	Files[0]=Files[1]=Files[2]="";
	GraphT=new TGraph(Disp);
	GraphT->Fit=0; GraphT->XLPos=2; GraphT->YLPos=4;
	for (int i=0;i<3;i++) {
		GraphG[i]=new TGraph(Disp);
		if (i<2) GraphG[i]->XLPos=0;
	}
	for (int i=0;i<3;i++) {
		OPos[i]=0.0; for (int j=0;j<3;j++) RPos[j][i]=0.0;
	}
	PlotType->ItemIndex=PlotTypeD;
	QFlag->ItemIndex=QFlagD;
	DoubleBuffered=true;
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::FormShow(TObject *Sender)
{
	UpdateColor();
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::BtnCloseClick(TObject *Sender)
{
	PlotTypeD=PlotType->ItemIndex; QFlagD=QFlag->ItemIndex;
	Release();
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::DispPaint(TObject *Sender)
{
	TCanvas *c=Disp->Canvas;
	c->Pen->Color=CColor[0]; c->Pen->Style=psSolid;
	c->Brush->Color=CColor[0]; c->Brush->Style=bsSolid;
	c->FillRect(Disp->ClientRect);
	switch (PlotType->ItemIndex) {
	case 0: DrawGraphT(Drag?0:1); break;
	case 1: DrawGraphG(Drag?0:1); break;
	}
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::Refresh(void)
{
	Disp->Invalidate();
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::DispMouseDown(TObject *Sender, TMouseButton Button,
      TShiftState Shift, int X, int Y)
{
	X0=X; Y0=Y;
	Drag=Shift.Contains(ssLeft)?1:(Shift.Contains(ssRight)?11:0);
	
	if (PlotType->ItemIndex==0) {
		GraphT->GetCent(Xc,Yc);
		GraphT->GetScale(Xs,Ys);
		Screen->Cursor=Drag==1?crSizeAll:crSizeNS;
	}
	else if (PlotType->ItemIndex==1) {
		int area;
		TPoint p(X,Y);
		for (int i=0;i<3;i++) {
			GraphG[i]->GetCent(Xc,Yc);
			GraphG[i]->GetScale(Xs,Ys);
			area=GraphG[i]->OnAxis(p);
			if (Drag==1&&area==0) {Drag+=i; Screen->Cursor=crSizeAll; return;}
			else if (area==1) {Drag+=i+4; Screen->Cursor=crSizeNS; return;}
			else if (area==0) break;
		}
		if (area==0||area==8) {Drag+=3; Screen->Cursor=crSizeWE;} else Drag=0;
	}
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::DispMouseLeave(TObject *Sender)
{
	Xn=Yn=-1;
	Panel22->Visible=false;
	Message2->Caption="";
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::DispMouseMove(TObject *Sender, TShiftState Shift,
      int X, int Y)
{
	if (X!=Xn||Y!=Yn) {Xn=X; Yn=Y;} else return;
	
	double dx=(X0-X)*Xs,dy=(Y-Y0)*Ys;
	double dxs=pow(2.0,(X0-X)/100.0),dys=pow(2.0,(Y-Y0)/100.0);
	
	if (Drag==0) {
		UpdatePoint(X,Y);
		return;
	}
	else if (PlotType->ItemIndex==0) {
		if (Drag==1) {
			GraphT->SetCent(Xc+dx,Yc+dy);
		}
		else {
			GraphT->SetScale(Xs*dys,Ys*dys);
		}
	}
	else if (PlotType->ItemIndex==1) {
		if (Drag<=4) {
			for (int i=0;i<3;i++) {
				double x,y;
				GraphG[i]->GetCent(x,y);
				GraphG[i]->SetCent(Xc+dx,i==Drag-1?Yc+dy:y);
			}
		}
		else if (Drag<=7) {
			double x,y;
			GraphG[Drag-5]->GetCent(x,y);
			GraphG[Drag-5]->SetCent(x,Yc+dy);
		}
		else if (Drag<=14) {
			for (int i=0;i<3;i++) {
				double xs,ys;
				GraphG[i]->GetScale(xs,ys);
				GraphG[i]->SetScale(Xs*dxs,ys);
			}
		}
		else if (Drag<=17) {
			double xs,ys;
			GraphG[Drag-15]->GetScale(xs,ys);
			GraphG[Drag-15]->SetScale(xs,Ys*dys);
		}
	}
	Refresh();
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::DispMouseUp(TObject *Sender, TMouseButton Button,
      TShiftState Shift, int X, int Y)
{
	Drag=0; Screen->Cursor=crDefault;
	Refresh();
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::FormMouseWheel(TObject *Sender, TShiftState Shift,
      int WheelDelta, TPoint &MousePos, bool &Handled)
{
	double ds=pow(2.0,-WheelDelta/1200.0),xs,ys;
	Handled=true;
	if (Xn<0||Yn<0) return;
	if (PlotType->ItemIndex==0) {
		GraphT->GetScale(xs,ys);
		GraphT->SetScale(xs*ds,ys*ds);
		Refresh();
	}
	else if (PlotType->ItemIndex==1) {
		TPoint p(Xn,Yn);
		int area;
		for (int i=0;i<3;i++) {
			area=GraphG[i]->OnAxis(p);
			if (area==1||area==2) {
				GraphG[i]->GetScale(xs,ys);
				GraphG[i]->SetScale(xs,ys*ds);
			}
			else if (area==0) break;
		}
		if (area==0||area==8) {
			for (int i=0;i<3;i++) {
				GraphG[i]->GetScale(xs,ys);
				GraphG[i]->SetScale(xs*ds,ys);
			}
		}
		Refresh();
	}
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::QFlagChange(TObject *Sender)
{
	Refresh();
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::TimeScrollChange(TObject *Sender)
{
	UpdateTime();
	Refresh();
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::FormResize(TObject *Sender)
{
	int w=Disp->Width,h=Disp->Height,hh=(h-21)/3;
	TPoint p1(0,0),p2(w-1,h-1);
	GraphT->SetPos(p1,p2);
	p1.x+=35; p1.y+=5; p2.x-=5; p2.y=p1.y+hh;
	for (int i=0;i<3;i++,p1.y+=hh,p2.y+=hh) GraphG[i]->SetPos(p1,p2);
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::SelectChange(TObject *Sender)
{
	UpdateSelect();
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::PlotTypeChange(TObject *Sender)
{
	Refresh();
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::BtnOptionClick(TObject *Sender)
{
	int w=PlotOptDialog->Width,h=PlotOptDialog->Height;
	PlotOptDialog->Left=Left+Width/2-w/2;
	PlotOptDialog->Top=Top+Height/2-h/2;
	if (PlotOptDialog->ShowModal()!=mrOk) return;
	UpdateColor();
	UpdateOrigin();
	UpdateInfo();
	UpdateLim();
	Refresh();
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::Read(AnsiString file)
{
	char sep[32]=" ";
	int posf=MainForm->SolFormat;
	int degf=MainForm->LatLonFormat;
	if (*MainForm->FieldSep.c_str()) strcpy(sep,MainForm->FieldSep.c_str());
	int n,sel=Select->ItemIndex<3?Select->ItemIndex:0;
	if ((n=readsol(file.c_str(),posf,degf,0.0,0,sep,POSMAX,Time[sel],Pos[sel],
		RPos[sel]))<=0) {
		Message1->Caption="no position data : "+file;
		return;
	}
	NPos[sel]=n; Files[sel]=file;
	RPos[sel][0]*=D2R; RPos[sel][1]*=D2R;
	for (int i=0;i<n;i++) {
		Pos[sel][i][0]*=D2R; Pos[sel][i][1]*=D2R;
	}
	if (sel==0) {
		UpdateOrigin();
		UpdateLim();
	}
	Caption=Files[0]+" "+Files[1]+" "+Files[2];
	UpdateSelect();
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::DrawGraphT(int level)
{
	TPoint p1,p2;
	double xt,yt,xs,ys;
	GraphT->DrawAxis();
	GraphT->GetTick(xt,yt);
	GraphT->GetScale(xs,ys);
	GraphT->GetPos(p1,p2);
	p1.x+=40; p1.y+=60; p2.x-=70; p2.y-=25;
	GraphT->DrawMark(p1,13,CColor[2],50,0);
	GraphT->DrawMark(p2,11,CColor[2],(int)(xt/xs+0.5),0);
	p2.y-=5;
	GraphT->DrawText(p2,GraphT->NumText(xt,xt)+" m",CColor[2],0,1,0);
	for (int sel=0;sel<3;sel++) {
		int n,m=0;
		if ((n=NPos[sel])<=0) continue;
		if ((Select->ItemIndex<3&&Select->ItemIndex!=sel)) continue;
		if ((Select->ItemIndex>2&&Select->ItemIndex-2<sel)) continue;
		
		double *x=new double[n],*y=new double[n],*z=new double[n];
		SolToXyz(Pos[sel],n,x,y,z);
		if (level&&ShowArrow) DrawArrow(Time[sel],x,y,n);
		int *q=new int[n];
		for (int i=0;i<n;i++) {
			q[m]=(int)Pos[sel][i][3];
			if (QFlag->ItemIndex>0&&q[m]!=QFlag->ItemIndex) continue;
			x[m]=x[i]; y[m]=y[i]; z[m++]=z[i];
		}
		if (!level||!(PlotStyle%2)) GraphT->DrawPoly(x,y,m,CColor[3],sel);
		if (level&&PlotStyle<2) {
			for (int i=0;i<m;i++) {
				GraphT->DrawMark(x[i],y[i],sel,MColor[q[i]],MarkSize,0);
			}
		}
		if (ShowStats) {
			double ave[3],std[3],rms[3],r[3],rr[3],pos[3];
			CalcStats(x,m,0.0,ave[0],std[0],rms[0]);
			CalcStats(y,m,0.0,ave[1],std[1],rms[1]);
			CalcStats(z,m,0.0,ave[2],std[2],rms[2]);
			pos2ecef(OPos,r); enu2ecef(OPos,ave,rr);
			for (int i=0;i<3;i++) r[i]+=rr[i];
			ecef2pos(r,pos);
			AnsiString s[6];
			s[0].sprintf("ORI=%12.8f%c %13.8f%c %8.3fm",OPos[0]*R2D,CHARDEG,
					   OPos[1]*R2D,CHARDEG,OPos[2]);
			s[1].sprintf("AVE=%12.8f%c %13.8f%c %8.3fm",pos[0]*R2D,CHARDEG,
					   pos[1]*R2D,CHARDEG,pos[2]);
			s[2].sprintf("STD=E:%6.3fm N:%6.3fm U:%6.3fm",std[0],std[1],std[2]);
			s[3].sprintf("RMS=E:%6.3fm N:%6.3fm U:%6.3fm",rms[0],rms[1],rms[2]);
			TPoint p1,p2;
			GraphT->GetPos(p1,p2); p1.x=p2.x-10; p1.y+=8;
			for (int i=0;i<4;i++,p1.y+=14) {
				GraphT->DrawText(p1,s[i],CColor[2],2,2,0);
			}
		}
		delete [] x; delete [] y; delete [] z; delete [] q;
		
		double xx,yy,zz;
		int i=TimeScroll->Position;
		if (0<i&&i<n) {
			SolToXyz(&Pos[sel][i],1,&xx,&yy,&zz);
			TColor color=MColor[(int)Pos[sel][i][3]];
			GraphT->DrawMark(xx,yy,0,color,MarkSize*2,0);
			GraphT->DrawMark(xx,yy,1,CColor[2],MarkSize*2+6,0);
		}
		SolToXyz(&RPos[sel],1,&xx,&yy,&zz);
		GraphT->DrawMark(xx,yy,0,CColor[2],MarkSize,0);
		GraphT->DrawText(xx,yy,"Ref Pos",CColor[2],0,2,0);
	}
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::DrawGraphG(int level)
{
	AnsiString label[]={"E-W (m)","N-S (m)","U-D (m)"};
	for (int i=0;i<3;i++) {
		GraphG[i]->XLPos=TimeLabel?(i==2?6:5):(i==2?1:0);
		GraphG[i]->DrawAxis();
		TPoint p1,p2;
		GraphG[i]->GetPos(p1,p2); p1.x+=5; p1.y+=3;
		GraphG[i]->DrawText(p1,label[i],CColor[2],1,2,0);
	}
	int sel=Select->ItemIndex<3?Select->ItemIndex:0,n=NPos[sel],m=0;
	if (n<=0) return;
	
	double *t=new double[n],*x=new double[n],*y=new double[n],*z=new double[n];
	double *s[]={x,y,z};
	int *q=new int[n];
	SolToXyz(Pos[sel],n,x,y,z);
	for (int i=0;i<n;i++) {
		q[m]=(int)Pos[sel][i][3];
		if (QFlag->ItemIndex>0&&q[m]!=QFlag->ItemIndex) continue;
		t[m]=time2gpst(Time[sel][i],NULL); x[m]=x[i]; y[m]=y[i]; z[m++]=z[i];
	}
	for (int i=0;i<3;i++) {
		if (!level||!(PlotStyle%2)) GraphG[i]->DrawPoly(t,s[i],m,CColor[3],0);
		if (level&&PlotStyle<2) {
			for (int j=0;j<m;j++) {
				GraphG[i]->DrawMark(t[j],s[i][j],sel,MColor[q[j]],MarkSize,0);
			}
		}
		if (ShowStats) {
			double ave,std,rms;
			CalcStats(s[i],m,0.0,ave,std,rms);
			AnsiString str;
			TPoint p1,p2;
			GraphG[i]->GetPos(p1,p2); p1.x=p2.x-5; p1.y+=3;
			if (i==0) {
				str.sprintf("ORI=%12.8f%c %13.8f%c %8.3fm",OPos[0]*R2D,CHARDEG,
						   OPos[1]*R2D,CHARDEG,OPos[2]);
				GraphG[i]->DrawText(p1,str,CColor[2],2,2,0);
				p1.y+=14;
			}
			str.sprintf("AVE=%.3fm STD=%.3fm RMS=%.3fm",ave,std,rms);
			GraphG[i]->DrawText(p1,str,CColor[2],2,2,0);
		}
	}
	int i=TimeScroll->Position;
	if (0<i&&i<n) {
		double tt=time2gpst(Time[sel][i],NULL),xx,yy,zz;
		for (int j=0;j<3;j++) {
			double xl[2],yl[2];
			GraphG[j]->GetLim(xl,yl);
			xl[0]=xl[1]=tt;
			GraphG[j]->DrawPoly(xl,yl,2,CColor[2],0);
		}
		TColor color=MColor[(int)Pos[sel][i][3]];
		SolToXyz(&Pos[sel][i],1,&xx,&yy,&zz);
		GraphG[0]->DrawMark(tt,xx,0,color,MarkSize*2,0);
		GraphG[1]->DrawMark(tt,yy,0,color,MarkSize*2,0);
		GraphG[2]->DrawMark(tt,zz,0,color,MarkSize*2,0);
		GraphG[0]->DrawMark(tt,xx,1,CColor[2],MarkSize*2+6,0);
		GraphG[1]->DrawMark(tt,yy,1,CColor[2],MarkSize*2+6,0);
		GraphG[2]->DrawMark(tt,zz,1,CColor[2],MarkSize*2+6,0);
	}
	delete [] t; delete [] x; delete [] y; delete [] z; delete [] q;
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::UpdateSelect(void)
{
	int sel=Select->ItemIndex<3?Select->ItemIndex:0;
	TimeScroll->Max=NPos[sel]<=0?100:NPos[sel]-1;
	TimeScroll->Position=0;
	UpdateTime();
	UpdateInfo();
	Refresh();
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::UpdateTime(void)
{
	int sel=Select->ItemIndex<3?Select->ItemIndex:0,n=NPos[sel];
	if (n<=0) return;
	int i=TimeScroll->Position;
	i=i<n?i:n-1;
	char s[128];
	sprintf(s," : %12.8f%c %13.8f%c %8.3fm",Pos[sel][i][0]*R2D,CHARDEG,
			Pos[sel][i][1]*R2D,CHARDEG,Pos[sel][i][2]);
	if (0<i) {
		double vel,dir;
		vel=SolToVel(Time[sel]+i,Pos[sel][i],dir);
		sprintf(s+strlen(s)," v=%4.1fm/s (%3.0f%c)",vel,dir,CHARDEG);
	}
	Message1->Caption=TimeStr(Time[sel][i],2)+s;
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::UpdateInfo(void)
{
	int sel=Select->ItemIndex<3?Select->ItemIndex:0,n=NPos[sel],nq[8]={0};
	AnsiString s1="",s2="",s3="",s;
	if (n>0) {
		double r[3],rr[3],b,bl[2]={1E6,0.0};
		pos2ecef(RPos[sel],rr);
		s1=TimeStr(Time[sel][0],0);
		s2=TimeStr(Time[sel][n-1],0);
		if (TimeLabel) s2=s2.SubString(12,8);
		s3.sprintf("n=%d Q=",n);
		for (int i=0;i<n;i++) {
			pos2ecef(Pos[sel][i],r);
			for (int j=0;j<3;j++) r[j]-=rr[j];
			b=norm(r,3); if (b<bl[0]) bl[0]=b; if (b>bl[1]) bl[1]=b;
			nq[(int)Pos[sel][i][3]]++;
		}
		for (int i=0;i<8;i++) {
			if (nq[i]<=0) continue;
			s3+=s.sprintf("%d:%d(%.1f%%)",i,nq[i],(double)nq[i]/n*100.0);
		}
		s3+=s.sprintf(" B=%.1f-%.1fkm",bl[0]/1E3,bl[1]/1E3);
	}
	Message1->Caption=s1+"-"+s2+" : "+s3;
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::UpdatePoint(int x, int y)
{
	AnsiString str="";
	TPoint p(x,y);
	double e[3]={0},r[3],dr[3],pos[3],t,v;
	char s[64];
	switch (PlotType->ItemIndex) {
	case 0: if (OPos[0]==0.0&&OPos[1]==0.0) break;
			GraphT->ToPos(p,e[0],e[1]);
			pos2ecef(OPos,r); enu2ecef(OPos,e,dr);
			for (int i=0;i<3;i++) r[i]+=dr[i];
			ecef2pos(r,pos);
			str.sprintf("%12.8f%c %13.8f%c",pos[0]*R2D,CHARDEG,pos[1]*R2D,
				        CHARDEG); break;
	case 1: GraphG[0]->ToPos(p,t,v);
			int week; (void)time2gpst(Time[0][0],&week);
			str=TimeStr(gpst2time(week,t),2);
			break;
	}
	Panel22->Visible=true;
	Message2->Caption=str;
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::UpdateOrigin(void)
{
	int n=NPos[0];
	if (n<=0) return;
	
	double opos[3]={0};
	switch (Origin) {
	case 0: for (int i=0;i<3;i++) opos[i]=Pos[0][0][i]; break;
	case 1: for (int i=0;i<3;i++) opos[i]=Pos[0][n-1][i]; break;
	case 2:
		for (int i=0;i<3;i++) {
			for (int j=0;j<n;j++) opos[i]+=Pos[0][j][i];
			opos[i]/=n;
		}
		break;
	case 3: for (int i=0;i<3;i++) opos[i]=RPos[0][i]; break;
	case 4: for (int i=0;i<3;i++) opos[i]=OOPos[i]; break;
	}
	if (YRange<=0) {
		double rr[3],r[3],e[3],x,y;
		pos2ecef(OPos,rr); pos2ecef(opos,r);
		for (int i=0;i<3;i++) r[i]-=rr[i];
		ecef2enu(OPos,r,e);
		GraphT->GetCent(x,y); GraphT->SetCent(x-e[0],y-e[1]);
		for (int i=0;i<3;i++) {
			GraphG[i]->GetCent(x,y); GraphG[i]->SetCent(x,y-e[i]);
		}
	}
	for (int i=0;i<3;i++) OPos[i]=opos[i];
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::UpdateColor(void)
{
	for (int i=0;i<3;i++) GraphT->Color[i]=CColor[i];
	for (int i=0;i<3;i++) {
		for (int j=0;j<3;j++) GraphG[j]->Color[i]=CColor[i];
	}
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::UpdateLim(void)
{
	int sel=Select->ItemIndex<3?Select->ItemIndex:0,n=NPos[sel];
	if (n<=0) return;
	double tl[]={time2gpst(Time[sel][0],NULL),time2gpst(Time[sel][n-1],NULL)};
	double xl[]={1E8,-1E8},yl[2]={1E8,-1E8},zl[2]={1E8,-1E8};
	if (tl[1]<tl[0]) {double tmp; tmp=tl[0]; tl[0]=tl[1]; tl[1]=tmp;}
	if (AutoScale) {
		double *x=new double[n],*y=new double[n],*z=new double[n];
		SolToXyz(Pos[sel],n,x,y,z);
		for (int i=0;i<n;i++) {
			if (x[i]<xl[0]) xl[0]=x[i]; else if (xl[1]<x[i]) xl[1]=x[i];
			if (y[i]<yl[0]) yl[0]=y[i]; else if (yl[1]<y[i]) yl[1]=y[i];
			if (z[i]<zl[0]) zl[0]=z[i]; else if (zl[1]<z[i]) zl[1]=z[i];
		}
		delete [] x; delete [] y; delete [] z;
		xl[0]-=0.05; xl[1]+=0.05;
		yl[0]-=0.05; yl[1]+=0.05;
		zl[0]-=0.05; zl[1]+=0.05;
	}
	else {
		xl[0]=-YRange; xl[1]=YRange;
		yl[0]=-YRange; yl[1]=YRange;
		zl[0]=-YRange; zl[1]=YRange;
	}
	double xs,ys;
	GraphT->SetLim(xl,yl);
	GraphT->GetScale(xs,ys);
	GraphT->SetScale(xs>ys?xs:ys,xs>ys?xs:ys);
	GraphG[0]->SetLim(tl,xl);
	GraphG[1]->SetLim(tl,yl);
	GraphG[2]->SetLim(tl,zl);
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::SolToXyz(const double pos[][NSOL], int n,
	double x[], double y[], double z[])
{
	double rr[3],r[3],e[3];
	pos2ecef(OPos,rr);
	for (int i=0;i<n;i++) {
		pos2ecef(pos[i],r);
		for (int j=0;j<3;j++) r[j]-=rr[j];
		ecef2enu(OPos,r,e);
		x[i]=e[0]; y[i]=e[1]; z[i]=e[2];
	}
}
//---------------------------------------------------------------------------
double __fastcall TSolPlot::SolToVel(const gtime_t *time, const double *pos,
	double &dir)
{
	double d,v,r0[3],r1[3],e[3];
	dir=0.0;
	pos2ecef(pos-4,r0);
	pos2ecef(pos,r1);
	for (int i=0;i<3;i++) r1[i]-=r0[i];
	ecef2enu(pos,r1,e);
	if ((d=norm(e,2))<1E-3) return 0.0;
	if ((v=d/timediff(time[0],time[-1]))<0.05) return 0.0;
	dir=atan2(e[0],e[1])*R2D; if (dir<0.0) dir+=360.0;
	return v;
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::DrawArrow(const gtime_t *t, const double *x,
	const double *y, int n)
{
	for (int i=1;i<n-1;i++) {
		int off=8;
		double tt=time2gpst(t[i],NULL);
		double d[]={x[i+1]-x[i-1],y[i+1]-y[i-1]},dist=norm(d,2);
		double vel=dist/timediff(t[i+1],t[i-1]);
		if (vel<0.5||fmod(tt+0.005,INTARROW)>=0.01) continue;
		TPoint p;
		GraphT->ToPoint(x[i],y[i],p);
		p.x-=(int)(off*d[1]/dist); p.y-=(int)(off*d[0]/dist);
		GraphT->DrawMark(p,10,CColor[2],15,(int)(atan2(d[1],d[0])*R2D));
	}
}
//---------------------------------------------------------------------------
void __fastcall TSolPlot::CalcStats(const double *x, int n,
	double ref, double &ave, double &std, double &rms)
{
	if (n<=0) {
		ave=std=rms=0.0;
		return;
	}
	ave=std=rms=0.0;
	double sum=0.0,sumsq=0.0;
	for (int i=0;i<n;i++) {
		sum+=x[i]; sumsq+=x[i]*x[i];
	}
	ave=sum/n;
	std=n>1?sqrt((sumsq-2.0*sum*ave+ave*ave*n)/(n-1)):0.0;
	rms=sqrt((sumsq-2.0*sum*ref+ref*ref*n)/n);
}
//---------------------------------------------------------------------------
AnsiString __fastcall TSolPlot::TimeStr(gtime_t time, int n)
{
	char s[64];
	if (TimeLabel) time2str(time,s,n);
	else {
		int week; double t=time2gpst(time,&week);
		sprintf(s,"%4d/%*.*fs",week,(n>0?6:5)+n,n,t);
	}
	AnsiString str(s);
	return str;
}
//---------------------------------------------------------------------------
